@@ -30,9 +30,219 @@
 #include "../configuration.h"
 #include "vlcmediawidget.h"
 
-VlcMediaWidget::VlcMediaWidget(QWidget *parent) : AbstractMediaWidget(parent), vlcInstance(NULL),
-	vlcMediaPlayer(NULL), playingDvd(false)
+const char *vlcEventName(int event)
 {
+	switch (event) {
+	case libvlc_MediaMetaChanged:
+		return "MediaMetaChanged";
+	case libvlc_MediaPlayerEncounteredError:
+		return "MediaPlayerEncounteredError";
+	case libvlc_MediaPlayerEndReached:
+		return "MediaPlayerEndReached";
+	case libvlc_MediaPlayerLengthChanged:
+		return "MediaPlayerLengthChanged";
+	case libvlc_MediaPlayerSeekableChanged:
+		return "MediaPlayerSeekableChanged";
+	case libvlc_MediaPlayerStopped:
+		return "MediaPlayerStopped";
+	case libvlc_MediaPlayerTimeChanged:
+		return "MediaPlayerTimeChanged";
+	case libvlc_MediaSubItemAdded:
+		return "MediaSubItemAdded";
+	case libvlc_MediaDurationChanged:
+		return "MediaDurationChanged";
+	case libvlc_MediaParsedChanged:
+		return "MediaParsedChanged";
+	case libvlc_MediaFreed:
+		return "MediaFreed";
+	case libvlc_MediaStateChanged:
+		return "MediaStateChanged";
+	case libvlc_MediaSubItemTreeAdded:
+		return "MediaSubItemTreeAdded";
+	case libvlc_MediaPlayerMediaChanged:
+		return "MediaPlayerMediaChanged";
+	case libvlc_MediaPlayerNothingSpecial:
+		return "MediaPlayerNothingSpecial";
+	case libvlc_MediaPlayerOpening:
+		return "MediaPlayerOpening";
+	case libvlc_MediaPlayerBuffering:
+		return "MediaPlayerBuffering";
+	case libvlc_MediaPlayerPlaying:
+		return "MediaPlayerPlaying";
+	case libvlc_MediaPlayerPaused:
+		return "MediaPlayerPaused";
+	case libvlc_MediaPlayerForward:
+		return "MediaPlayerForward";
+	case libvlc_MediaPlayerBackward:
+		return "MediaPlayerBackward";
+	case libvlc_MediaPlayerPositionChanged:
+		return "MediaPlayerPositionChanged";
+	case libvlc_MediaPlayerPausableChanged:
+		return "MediaPlayerPausableChanged";
+	case libvlc_MediaPlayerTitleChanged:
+		return "MediaPlayerTitleChanged";
+	case libvlc_MediaPlayerSnapshotTaken:
+		return "MediaPlayerSnapshotTaken";
+	case libvlc_MediaPlayerVout:
+		return "MediaPlayerVout";
+	case libvlc_MediaPlayerScrambledChanged:
+		return "MediaPlayerScrambledChanged";
+	case libvlc_MediaPlayerUncorked:
+		return "MediaPlayerUncorked";
+	case libvlc_MediaPlayerMuted:
+		return "MediaPlayerMuted";
+	case libvlc_MediaListItemAdded:
+		return "MediaListItemAdded";
+	case libvlc_MediaListWillAddItem:
+		return "MediaListWillAddItem";
+	case libvlc_MediaListItemDeleted:
+		return "MediaListItemDeleted";
+	case libvlc_MediaListWillDeleteItem:
+		return "MediaListWillDeleteItem";
+	case libvlc_MediaListViewItemAdded:
+		return "MediaListViewItemAdded";
+	case libvlc_MediaListViewWillAddItem:
+		return "MediaListViewWillAddItem";
+	case libvlc_MediaListViewItemDeleted:
+		return "MediaListViewItemDeleted";
+	case libvlc_MediaListViewWillDeleteItem:
+		return "MediaListViewWillDeleteItem";
+	case libvlc_MediaListPlayerPlayed:
+		return "MediaListPlayerPlayed";
+	case libvlc_MediaListPlayerNextItemSet:
+		return "MediaListPlayerNextItemSet";
+	case libvlc_MediaListPlayerStopped:
+		return "MediaListPlayerStopped";
+	case libvlc_RendererDiscovererItemAdded:
+		return "RendererDiscovererItemAdded";
+	case libvlc_RendererDiscovererItemDeleted:
+		return "RendererDiscovererItemDeleted";
+	case libvlc_VlmMediaAdded:
+		return "VlmMediaAdded";
+	case libvlc_VlmMediaRemoved:
+		return "VlmMediaRemoved";
+	case libvlc_VlmMediaChanged:
+		return "VlmMediaChanged";
+	case libvlc_VlmMediaInstanceStarted:
+		return "VlmMediaInstanceStarted";
+	case libvlc_VlmMediaInstanceStopped:
+		return "VlmMediaInstanceStopped";
+	case libvlc_VlmMediaInstanceStatusInit:
+		return "VlmMediaInstanceStatusInit";
+	case libvlc_VlmMediaInstanceStatusOpening:
+		return "VlmMediaInstanceStatusOpening";
+	case libvlc_VlmMediaInstanceStatusPlaying:
+		return "VlmMediaInstanceStatusPlaying";
+	case libvlc_VlmMediaInstanceStatusPause:
+		return "VlmMediaInstanceStatusPause";
+	case libvlc_VlmMediaInstanceStatusEnd:
+		return "VlmMediaInstanceStatusEnd";
+	case libvlc_VlmMediaInstanceStatusError:
+		return "VlmMediaInstanceStatusError";
+#if LIBVLC_VERSION_MAJOR > 2
+	case libvlc_MediaPlayerAudioVolume:
+		return "MediaPlayerAudioVolume";
+	case libvlc_MediaPlayerAudioDevice:
+		return "MediaPlayerAudioDevice";
+	case libvlc_MediaListEndReached:
+		return "MediaListEndReached";
+	case libvlc_MediaPlayerChapterChanged:
+		return "MediaPlayerChapterChanged";
+	case libvlc_MediaPlayerESAdded:
+		return "MediaPlayerESAdded";
+	case libvlc_MediaPlayerESDeleted:
+		return "MediaPlayerESDeleted";
+	case libvlc_MediaPlayerESSelected:
+		return "MediaPlayerESSelected";
+	case libvlc_MediaPlayerCorked:
+		return "MediaPlayerCorked";
+	case libvlc_MediaPlayerUnmuted:
+		return "MediaPlayerUnmuted";
+#endif
+
+	default:
+		return "Unknown";
+	}
+}
+
+VlcMediaWidget::VlcMediaWidget(QWidget *parent) : AbstractMediaWidget(parent),
+    timer(NULL), vlcInstance(NULL), vlcMedia(NULL), vlcMediaPlayer(NULL),
+    isPaused(false), playingDvd(false), urlIsAudioCd(false),
+    typeOfDevice(""), trackNumber(1), numTracks(1)
+{
+	libvlc_event_e events[] = {
+		libvlc_MediaMetaChanged,
+		libvlc_MediaPlayerEncounteredError,
+		libvlc_MediaPlayerEndReached,
+		libvlc_MediaPlayerLengthChanged,
+		libvlc_MediaPlayerSeekableChanged,
+		libvlc_MediaPlayerStopped,
+#if LIBVLC_VERSION_MAJOR > 2
+		libvlc_MediaPlayerESAdded,
+		libvlc_MediaPlayerESDeleted,
+#endif
+		libvlc_MediaPlayerTimeChanged,
+#if 0 // all other possible events
+		libvlc_MediaSubItemAdded,
+		libvlc_MediaDurationChanged,
+		libvlc_MediaParsedChanged,
+		libvlc_MediaFreed,
+		libvlc_MediaStateChanged,
+		libvlc_MediaSubItemTreeAdded,
+		libvlc_MediaPlayerMediaChanged,
+		libvlc_MediaPlayerNothingSpecial,
+		libvlc_MediaPlayerOpening,
+		libvlc_MediaPlayerBuffering,
+		libvlc_MediaPlayerPlaying,
+		libvlc_MediaPlayerPaused,
+		libvlc_MediaPlayerForward,
+		libvlc_MediaPlayerBackward,
+		libvlc_MediaPlayerPositionChanged,
+		libvlc_MediaPlayerPausableChanged,
+		libvlc_MediaPlayerTitleChanged,
+		libvlc_MediaPlayerSnapshotTaken,
+		libvlc_MediaPlayerVout,
+		libvlc_MediaPlayerScrambledChanged,
+		libvlc_MediaPlayerUncorked,
+		libvlc_MediaPlayerMuted,
+		libvlc_MediaListItemAdded,
+		libvlc_MediaListWillAddItem,
+		libvlc_MediaListItemDeleted,
+		libvlc_MediaListWillDeleteItem,
+		libvlc_MediaListViewItemAdded,
+		libvlc_MediaListViewWillAddItem,
+		libvlc_MediaListViewItemDeleted,
+		libvlc_MediaListViewWillDeleteItem,
+		libvlc_MediaListPlayerPlayed,
+		libvlc_MediaListPlayerNextItemSet,
+		libvlc_MediaListPlayerStopped,
+		libvlc_VlmMediaAdded,
+		libvlc_VlmMediaRemoved,
+		libvlc_VlmMediaChanged,
+		libvlc_VlmMediaInstanceStarted,
+		libvlc_VlmMediaInstanceStopped,
+		libvlc_VlmMediaInstanceStatusInit,
+		libvlc_VlmMediaInstanceStatusOpening,
+		libvlc_VlmMediaInstanceStatusPlaying,
+		libvlc_VlmMediaInstanceStatusPause,
+		libvlc_VlmMediaInstanceStatusEnd,
+		libvlc_VlmMediaInstanceStatusError,
+#if LIBVLC_VERSION_MAJOR > 2
+		libvlc_MediaListEndReached,
+		libvlc_MediaPlayerAudioDevice,
+		libvlc_MediaPlayerAudioVolume,
+		libvlc_MediaPlayerChapterChanged,
+		libvlc_MediaPlayerESSelected,
+		libvlc_MediaPlayerCorked,
+		libvlc_MediaPlayerUnmuted,
+		libvlc_RendererDiscovererItemAdded,
+		libvlc_RendererDiscovererItemDeleted,
+#endif /*LIBVLC_VERSION_MAJOR */
+#endif
+	};
+
+	for (uint i = 0; i < (sizeof(events) / sizeof(events[0])); ++i)
+		eventType.append(events[i]);
 }
 
 bool VlcMediaWidget::init()
@@ -83,21 +293,10 @@ bool VlcMediaWidget::init()
 		return false;
 	}
 
-	libvlc_event_manager_t *eventManager = libvlc_media_player_event_manager(vlcMediaPlayer);
-	libvlc_event_e eventTypes[] = { libvlc_MediaPlayerEncounteredError,
-		libvlc_MediaPlayerEndReached, libvlc_MediaPlayerLengthChanged,
-		libvlc_MediaPlayerSeekableChanged, libvlc_MediaPlayerStopped,
-#if LIBVLC_VERSION_MAJOR > 2
-		libvlc_MediaPlayerESAdded, libvlc_MediaPlayerESDeleted,
-#endif
-		libvlc_MediaPlayerTimeChanged };
+	eventManager = libvlc_media_player_event_manager(vlcMediaPlayer);
 
-	for (uint i = 0; i < (sizeof(eventTypes) / sizeof(eventTypes[0])); ++i) {
-		if (libvlc_event_attach(eventManager, eventTypes[i], vlcEventHandler, this) != 0) {
-			qCCritical(logMediaWidget, "Cannot attach event handler %d", eventTypes[i]);
-			return false;
-		}
-	}
+	if (!registerEvents())
+		return false;
 
 	timer = new QTimer();
 	connect(timer, SIGNAL(timeout()), this, SLOT(hideMouse()));
@@ -116,13 +315,26 @@ bool VlcMediaWidget::init()
 
 VlcMediaWidget::~VlcMediaWidget()
 {
-	if (vlcMediaPlayer != NULL) {
-		libvlc_media_player_release(vlcMediaPlayer);
+	metadata.clear();
+	audioStreams.clear();
+	subtitles.clear();
+	subtitleId.clear();
+
+	if (timer != NULL) {
+		timer->stop();
+		delete timer;
 	}
 
-	if (vlcInstance != NULL) {
+	unregisterEvents();
+
+	if (vlcMedia != NULL)
+		libvlc_media_release(vlcMedia);
+
+	if (vlcInstance != NULL)
 		libvlc_release(vlcInstance);
-	}
+
+	if (vlcMediaPlayer != NULL)
+		libvlc_media_player_release(vlcMediaPlayer);
 }
 
 VlcMediaWidget *VlcMediaWidget::createVlcMediaWidget(QWidget *parent)
@@ -321,6 +533,11 @@ void VlcMediaWidget::play(const MediaSource &source)
 
 	typeOfDevice = url.constData();
 
+	if (vlcMedia != NULL) {
+		libvlc_media_player_stop(vlcMediaPlayer);
+		libvlc_media_release(vlcMedia);
+	}
+
 	vlcMedia = libvlc_media_new_location(vlcInstance, typeOfDevice);
 	if (urlIsAudioCd)
 		libvlc_media_add_option(vlcMedia, "cdda-track=1");
@@ -336,6 +553,24 @@ void VlcMediaWidget::play(const MediaSource &source)
 	setMouseTracking(true);
 }
 
+void VlcMediaWidget::unregisterEvents()
+{
+	for (int i = 0; i < eventType.size(); ++i)
+		libvlc_event_detach(eventManager, eventType.at(i),
+				    vlcEventHandler, this);
+}
+
+bool VlcMediaWidget::registerEvents()
+{
+	for (int i = 0; i < eventType.size(); ++i) {
+		if (libvlc_event_attach(eventManager, eventType.at(i), vlcEventHandler, this) != 0) {
+			qCCritical(logMediaWidget, "Cannot attach event handler %d", eventType.at(i));
+			return false;
+		}
+	}
+	return true;
+}
+
 int VlcMediaWidget::makePlay()
 {
 	if (vlcMedia == NULL) {
@@ -343,18 +578,7 @@ int VlcMediaWidget::makePlay()
 		return -1;
 	}
 
-	libvlc_event_manager_t *eventManager = libvlc_media_event_manager(vlcMedia);
-	libvlc_event_e eventTypes[] = { libvlc_MediaMetaChanged };
-
-	for (uint i = 0; i < (sizeof(eventTypes) / sizeof(eventTypes[0])); ++i) {
-		if (libvlc_event_attach(eventManager, eventTypes[i], vlcEventHandler, this) != 0) {
-			qCWarning(logMediaWidget, "Cannot attach event handler %d", eventTypes[i]);
-		}
-	}
-
 	libvlc_media_player_set_media(vlcMediaPlayer, vlcMedia);
-	libvlc_media_release(vlcMedia);
-	vlcMedia = NULL;
 
 	/*
 	 * FIXME: This is mostly a boilerplate as, at least with vlc 3,
@@ -384,8 +608,10 @@ void VlcMediaWidget::playDirection(int direction)
 
 	strBuf += QString::number(trackNumber);
 
-	if (vlcMedia != NULL)
+	if (vlcMedia != NULL) {
+		libvlc_media_player_stop(vlcMediaPlayer);
 		libvlc_media_release(vlcMedia);
+	}
 
 	vlcMedia = libvlc_media_new_location(vlcInstance, typeOfDevice);
 	libvlc_media_add_option(vlcMedia, strBuf.toUtf8());
@@ -409,10 +635,16 @@ void VlcMediaWidget::stop()
 {
 	libvlc_media_player_stop(vlcMediaPlayer);
 
+	if (vlcMedia != NULL) {
+		libvlc_media_release(vlcMedia);
+		vlcMedia = NULL;
+	}
+
 	timer->stop();
 	setCursor(Qt::BlankCursor);
 	setCursor(Qt::ArrowCursor);
-	addPendingUpdates(PlaybackStatus);
+	addPendingUpdates(PlaybackStatus | Metadata |
+			  Subtitles | AudioStreams);
 }
 
 void VlcMediaWidget::setPaused(bool paused)
@@ -431,14 +663,20 @@ void VlcMediaWidget::seek(int time)
 	libvlc_media_player_set_time(vlcMediaPlayer, time);
 }
 
-void VlcMediaWidget::setCurrentAudioStream(int currentAudioStream)
+void VlcMediaWidget::setCurrentAudioStream(int _currentAudioStream)
 {
+	if (_currentAudioStream < 0 ||
+	    _currentAudioStream > libvlc_audio_get_track(vlcMediaPlayer))
+		_currentAudioStream = 0;
 	// skip the 'deactivate' audio channel
-	libvlc_audio_set_track(vlcMediaPlayer, currentAudioStream + 1);
+	libvlc_audio_set_track(vlcMediaPlayer, _currentAudioStream + 1);
+
+	currentAudioStream = _currentAudioStream;
 }
 
 void VlcMediaWidget::setCurrentSubtitle(int currentSubtitle)
 {
+	libvlc_track_description_t *tr, *track;
 	int requestedSubtitle = -1;
 
 	QMap<int, int>::const_iterator i = subtitleId.constBegin();
@@ -455,7 +693,9 @@ void VlcMediaWidget::setCurrentSubtitle(int currentSubtitle)
 	libvlc_video_set_spu(vlcMediaPlayer, requestedSubtitle);
 
 	/* Print what it was actually selected */
-	libvlc_track_description_t *track = libvlc_video_get_spu_description(vlcMediaPlayer);
+
+	tr = libvlc_video_get_spu_description(vlcMediaPlayer);
+	track = tr;
 	while (track != NULL) {
 		QString subtitle = QString::fromUtf8(track->psz_name);
 
@@ -467,7 +707,7 @@ void VlcMediaWidget::setCurrentSubtitle(int currentSubtitle)
 			qCDebug(logVlc, "Subtitle set to id %d: %s", track->i_id, qPrintable(subtitle));
 		track = track->p_next;
 	}
-	libvlc_track_description_list_release(track);
+	libvlc_track_description_list_release(tr);
 
 }
 
@@ -615,18 +855,31 @@ void VlcMediaWidget::updateSeekable()
 
 void VlcMediaWidget::updateMetadata()
 {
-	metadata.clear();
-	libvlc_media_t *media = libvlc_media_player_get_media(vlcMediaPlayer);
+	char *meta;
 
-	if (media != NULL) {
-		metadata.insert(MediaWidget::Title,
-			QString::fromUtf8(libvlc_media_get_meta(media, libvlc_meta_Title)));
-		metadata.insert(MediaWidget::Artist,
-			QString::fromUtf8(libvlc_media_get_meta(media, libvlc_meta_Artist)));
-		metadata.insert(MediaWidget::Album,
-			QString::fromUtf8(libvlc_media_get_meta(media, libvlc_meta_Album)));
-		metadata.insert(MediaWidget::TrackNumber,
-			QString::fromUtf8(libvlc_media_get_meta(media, libvlc_meta_TrackNumber)));
+	metadata.clear();
+	if (vlcMedia != NULL) {
+		QString s;
+		meta = libvlc_media_get_meta(vlcMedia, libvlc_meta_Title);
+		if (meta) {
+			metadata.insert(MediaWidget::Title, QString::fromUtf8(meta));
+			free(meta);
+		}
+		meta = libvlc_media_get_meta(vlcMedia, libvlc_meta_Artist);
+		if (meta) {
+			metadata.insert(MediaWidget::Artist, QString::fromUtf8(meta));
+			free(meta);
+		}
+		meta = libvlc_media_get_meta(vlcMedia, libvlc_meta_Album);
+		if (meta) {
+			metadata.insert(MediaWidget::Album, QString::fromUtf8(meta));
+			free(meta);
+		}
+		meta = libvlc_media_get_meta(vlcMedia, libvlc_meta_TrackNumber);
+		if (meta) {
+			metadata.insert(MediaWidget::TrackNumber, QString::fromUtf8(meta));
+			free(meta);
+		}
 	}
 
 	if (urlIsAudioCd)
@@ -637,7 +890,10 @@ void VlcMediaWidget::updateMetadata()
 void VlcMediaWidget::updateAudioStreams()
 {
 	audioStreams.clear();
-	libvlc_track_description_t *track = libvlc_audio_get_track_description(vlcMediaPlayer);
+	libvlc_track_description_t *tr, *track;
+
+	tr = libvlc_audio_get_track_description(vlcMediaPlayer);
+	track = tr;
 
 	if (track != NULL) {
 		// skip the 'deactivate' audio channel
@@ -664,18 +920,25 @@ void VlcMediaWidget::updateAudioStreams()
 		audioStreams.append(audioStream);
 		track = track->p_next;
 	}
+	libvlc_track_description_list_release(tr);
 
 	// skip the 'deactivate' audio channel
 	currentAudioStream = (libvlc_audio_get_track(vlcMediaPlayer) - 1);
+	if (currentAudioStream < 0)
+		setCurrentAudioStream(0);
 }
 
 void VlcMediaWidget::updateSubtitles()
 {
-	subtitles.clear();
-	libvlc_track_description_t *track = libvlc_video_get_spu_description(vlcMediaPlayer);
+	libvlc_track_description_t *tr, *track;
 	int i = 0;
 
+	subtitles.clear();
 	subtitleId.clear();
+
+	tr = libvlc_video_get_spu_description(vlcMediaPlayer);
+	track = tr;
+
 	if (track != NULL) {
 		// skip the 'deactivate' subtitle
 		track = track->p_next;
@@ -696,7 +959,7 @@ void VlcMediaWidget::updateSubtitles()
 		qCDebug(logVlc, "Got subtitle id#%d: %s", track->i_id, qPrintable(subtitle));
 		track = track->p_next;
 	}
-	libvlc_track_description_list_release(track);
+	libvlc_track_description_list_release(tr);
 
 	// skip the 'deactivate' subtitle
 	currentSubtitle = subtitleId.value(libvlc_video_get_spu(vlcMediaPlayer), -1);
@@ -792,7 +1055,7 @@ void VlcMediaWidget::vlcEvent(const libvlc_event_t *event)
 	case libvlc_MediaPlayerESDeleted:
 #endif
 	case libvlc_MediaMetaChanged:
-		pendingUpdatesToBeAdded = Metadata | Subtitles;
+		pendingUpdatesToBeAdded = Metadata | Subtitles | AudioStreams;
 		break;
 	case libvlc_MediaPlayerEncounteredError:
 		pendingUpdatesToBeAdded = PlaybackStatus;
@@ -818,7 +1081,8 @@ void VlcMediaWidget::vlcEvent(const libvlc_event_t *event)
 	if (pendingUpdatesToBeAdded != 0) {
 		addPendingUpdates(pendingUpdatesToBeAdded);
 	} else {
-		qCWarning(logMediaWidget, "Unknown libVLC event type %d", event->type);
+		qCWarning(logMediaWidget, "Unhandled event %s (%d)",
+			  vlcEventName(event->type), event->type);
 	}
 }
 
