@@ -290,7 +290,7 @@ void MainWindow::run()
 
 	// control bar
 
-	controlBar = new QToolBar(QLatin1String("control_bar"));
+	controlBar = new QToolBar(QLatin1String("control_bar"), this);
 	this->addToolBar(Qt::BottomToolBarArea, controlBar);
 
 	controlBar->setToolButtonStyle(Qt::ToolButtonIconOnly);
@@ -437,8 +437,13 @@ void MainWindow::close()
 	dvbTab->mayCloseApplication(&ok, mediaWidget);
 #endif /* HAVE_DVB == 1 */
 
-	if (ok)
+	if (ok) {
+		writeSettings();
+		mediaWidget->stop();
+		QMainWindow::close();
+		Configuration::detach();
 		QCoreApplication::exit(0);
+	}
 }
 
 void MainWindow::readSettings()
@@ -832,8 +837,7 @@ void MainWindow::closeEvent(QCloseEvent *event)
 	if (!ok) {
 		event->ignore();
 	} else {
-		writeSettings();
-		QMainWindow::closeEvent(event);
+		MainWindow::close();
 	}
 }
 
