@@ -24,6 +24,7 @@
 #include <KConfigGroup>
 #include <KSharedConfig>
 #include <KToolBar>
+#include <QAbstractItemView>
 #include <QBoxLayout>
 #include <QComboBox>
 #include <QContextMenuEvent>
@@ -793,14 +794,14 @@ void MediaWidget::previous()
 {
 	if (source->getType() == MediaSource::Url)
 		emit playlistPrevious();
-	backend->jumpToPreviousChapter();
+       source->previous();
 }
 
 void MediaWidget::next()
 {
 	if (source->getType() == MediaSource::Url)
 		emit playlistNext();
-	backend->jumpToNextChapter();
+       source->next();
 }
 
 void MediaWidget::stop()
@@ -1105,6 +1106,8 @@ void MediaWidget::currentAudioStreamChanged(int currentAudioStream)
 			backend->setCurrentAudioStream(currentAudioStream);
 		}
 	}
+	audioStreamBox->view()->setMinimumWidth(audioStreamBox->view()->sizeHintForColumn(0));
+
 }
 
 void MediaWidget::currentSubtitleChanged(int currentSubtitle)
